### Code accompanying the paper "Cryptographically secure multiparty evaluation
### of system reliability"

################################################################################
### THE PROTOCOL STARTS HERE.                                                ###
### This is the code to be run by the system designer.                       ###
################################################################################

# Load required libraries.
library("HomomorphicEncryption")
library("ReliabilityTheory")
source("utility.R")

# Specify the grid of times at which the system survival curve is desired.
t <- seq(0, 5, length.out=100)
# Also the precision desired during the computation
kappa <- 5

# Define the system and compute the survival signature.
ab <- graph.formula(s -- M -- C1:C2:C3:C4, P1:P2:P3:P4 -- t,
                    C1 -- P1, C2 -- P2, C3 -- P3, C4 -- P4, s -- H -- P3:P4)
ab <- setCompTypes(ab, list("C"=c("C1", "C2", "C3", "C4"),
                            "H"=c("H"),
                            "M"=c("M"),
                            "P"=c("P1", "P2", "P3", "P4")))

survsig <- computeSystemSurvivalSignature(ab)

# Store the lookup part of the table (l values) for each manufacturer.
XiL.C <- survsig$C
XiL.H <- survsig$H
XiL.M <- survsig$M
XiL.P <- survsig$P

# Extract out the probabilities for encryption.  These are stored separately
# from the l values unlike the paper.
# This involves duplicating the survival signature column as many times as 
# there are time points in the grid.
XiR.m <- matrix(survsig$Probability, nrow=nrow(survsig), ncol=length(t))

# Generate the encryption keys.
p <- parsHelp("FandV", L=8, max=4000, lambda=128) # request params supporting multiplicative depth 4.
k <- keygen(p)

# Encrypt the survival signature.
XiR.c <- enc(k$pk, mu(kappa, XiR.m))

# For safety the following which modify the filesystem have been commented out.
# Review the code to make sure you are happy and then uncomment and run.

# Save Xi then compress ... uncomment the following to run
# NB: XiR.c requires about 11.6GB storage uncompressed or 5.5GB compressed.
#saveFHE(XiR.c, "XiR_c.fhe")
#system("pigz XiR_c.fhe")

# Save public key ... uncomment the following to run
#saveFHE(k$pk, "pk.fhe")

# Save other parameters ... uncomment the following to run
#save(t, kappa, file="t_and_kappa.RData")
#save(XiL.C, file="XiL_C.RData")
#save(XiL.H, file="XiL_H.RData")
#save(XiL.M, file="XiL_M.RData")
#save(XiL.P, file="XiL_P.RData")

# Send public key, t, kappa and the appropriate XiL.? file to all manufacturers,
# for example over SSH tunnel ... modify the destintation, then uncomment the
# following to run
#system("scp pk.fhe username@manufacturer-c.com:~")
#system("scp t_and_kappa.RData username@manufacturer-c.com:~")
#system("scp XiL_C.RData username@manufacturer-c.com:~")
#system("scp pk.fhe username@manufacturer-h.com:~")
#system("scp t_and_kappa.RData username@manufacturer-h.com:~")
#system("scp XiL_H.RData username@manufacturer-h.com:~")
#system("scp pk.fhe username@manufacturer-m.com:~")
#system("scp t_and_kappa.RData username@manufacturer-m.com:~")
#system("scp XiL_M.RData username@manufacturer-m.com:~")
#system("scp pk.fhe username@manufacturer-p.com:~")
#system("scp t_and_kappa.RData username@manufacturer-p.com:~")
#system("scp XiL_P.RData username@manufacturer-p.com:~")

# Send the XiR.c to manufacturer C, for example over SSH tunnel ... modify the
# destintation, then uncomment the following to run
#system("scp XiR_c.fhe.gz username@manufacturer-c.com:~")

################################################################################
### The designer now sends the k$pk, t, and kappa to all manufacturers.      ###
### Further: XiL.C is sent to manufacturer C;                                ###
###          XiL.H is sent to manufacturer H;                                ###
###          XiL.M is sent to manufacturer M;                                ###
###          XiL.P is sent to manufacturer P.                                ###
### Finally, XiR.c is sent to the first manufacturer who will commence the   ###
### execution of the protocol, that is manufacturer C here.                  ###
###                                                                          ###
### Then, wait for the fourth manufacturer to return xi.c, the final answer  ###
### on the grid of times values before decrypting and plotting the survival  ###
### curve.                                                                   ###
################################################################################

# Load the results returned by manufacturer p
system("unpigz xi_c.fhe.gz")
xi.c <- loadFHE("xi_c.fhe")

# Now decrypt the updated survival signature.
xi.m <- dec(k$sk, xi.c)

# Reverse the integer encoding, accounting for the precision adjustment due to
# the 4 manufacturers
xi.m <- muinv(5*kappa, xi.m)

# Plot the survival curve
plot(t, xi.m, type="l", ylab="P(T_S>t)")
